open Printf
open Effect
open Effect.Deep

(** {1 Ordinary exceptions} *)

exception Conv_failure of string

let int_of_string s =
  match int_of_string_opt s with
  | Some n -> n
  | None -> raise (Conv_failure s)

let sum_stringlist l =
  l |> List.map int_of_string |> List.fold_left (+) 0

(*
let n0 = sum_stringlist ["1"; "xxx"; "2"]
*)

let safe_sum_stringlist lst =
  try
    sum_stringlist lst
  with Conv_failure s ->
    printf "Bad input: %s\n" s; max_int

(** {1 Resumable exceptions} *)

type _ Effect.t += Conversion_failure : string -> int Effect.t

let int_of_string s =
  match int_of_string_opt s with
  | Some n -> n
  | None -> perform (Conversion_failure s)

let sum_stringlist l =
  l |> List.map int_of_string |> List.fold_left (+) 0

(*
let n1 = sum_stringlist ["1"; "xxx"; "2"; "yyy"]
*)

let safe_sum_stringlist l =
  try_with sum_stringlist l
  { effc = (fun (type c) (eff: c Effect.t) ->
      match eff with
      | Conversion_failure s -> Some (fun (k: (c,_) continuation) ->
              printf "Bad input %s, replaced with 0\n" s;
              continue k 0)
      | _ -> None
    ) }

let n2 = safe_sum_stringlist ["1"; "xxx"; "2"; "yyy"]

(** Same example, with a shallow handler. *)

module M = struct

open Effect.Shallow

let safe1_sum_stringlist l =
  continue_with (fiber sum_stringlist) l
    { retc = (fun x -> x);
      effc = (fun (type c) (eff: c Effect.t) ->
        match eff with
        | Conversion_failure s -> Some (fun (k: (c,_) continuation) ->
            printf "This time I'll let it pass\n";
            continue_with k 0
              { retc = (fun x -> x);
                effc = (fun (type c) (eff: c Effect.t) -> None);
                exnc = (fun exn -> raise exn) })
        | _ -> None);
      exnc = (fun exn -> raise exn) }

let n3 = safe1_sum_stringlist ["1"; "xxx"; "2"]

(*
let n4 = safe1_sum_stringlist ["1"; "xxx"; 2; "yyy"]
*)

(** Exercise: implement a "three strike" approach, where up to 3 errors can
    be handled, with gradually more menacing messages. *)

end

(** {1 Iterators and generators} *)

(** Naive primality test. *)

let isprime n =
  let rec try_divide i =
    if i * i > n then true else
    if n mod i = 0 then false else
    try_divide (i + 2) in
  n mod 2 <> 0 && try_divide 3

(** Iterate [f] over all primes. *)

let iter_primes (f: int -> unit) : unit =
  for n = 2 to max_int do
    if isprime n then f n
  done

(*
let _ = iter_primes (fun n -> printf "%d\n" n)
*)

(** Produce the sequence of all primes. *)

let gen_primes : int Seq.t =
  let rec gen n : int Seq.t =
    if isprime n then (fun () -> Seq.Cons(n, gen (n + 1))) else gen (n + 1)
  in gen 2

let pp = Seq.drop 1000 gen_primes |> Seq.take 20 |> List.of_seq

(** Implementing [gen_primes] from [iter_prime], using control inversion. *)

type _ Effect.t += Next_prime : int -> unit Effect.t

let gen_primes : int Seq.t =
  match_with iter_primes (fun n -> perform (Next_prime n))
  { retc = (fun x -> Seq.empty);
    effc = (fun (type c) (eff: c Effect.t) ->
      match eff with
      | Next_prime n -> Some (fun (k: (c,_) continuation) ->
          fun () -> Seq.Cons(n, continue k ()))
      | _ -> None
    );
    exnc = (fun exn -> raise exn)
  }

let pp = Seq.take 20 gen_primes |> List.of_seq

(** More general transformation from iterator to sequence. *)

let iterator_to_sequence
       (type elt) (type collection)
       (iter: (elt -> unit) -> collection -> unit) : collection -> elt Seq.t =
  let module I2S = struct
    type _ Effect.t += Next : elt -> unit Effect.t
    let gen coll =
      match_with (iter (fun elt -> perform (Next elt))) coll
      { retc = (fun x -> Seq.empty);
        effc = (fun (type c) (eff: c Effect.t) ->
          match eff with
          | Next elt -> Some (fun (k: (c,_) continuation) ->
              fun () -> Seq.Cons(elt, continue k ()))
          | _ -> None
        );
        exnc = (fun exn -> raise exn)
      }
  end in I2S.gen

(** Application: the "same fringe" problem. *)

let same_fringe
    (iter1: ('elt -> unit) -> 'coll1 -> unit)
    (iter2: ('elt -> unit) -> 'coll2 -> unit)
    coll1 coll2 =
  Seq.for_all2 (=) (iterator_to_sequence iter1 coll1)
                   (iterator_to_sequence iter2 coll2)

module IntSet = Set.Make(Int)

let same =
  same_fringe List.iter IntSet.iter
              [1; 2; 3]
              (IntSet.add 2 (IntSet.add 1 (IntSet.singleton 3)))

(** {1 Coroutines} *)

type _ Effect.t +=
  | Spawn : (unit -> unit) -> unit Effect.t
  | Yield : unit Effect.t

let spawn f = perform (Spawn f)

let yield () = perform Yield

(** The queue of runnable tasks (stopped on [yield], ready to restart) *)

let runnable : (unit -> unit) Queue.t = Queue.create()

let suspend f = Queue.add f runnable

let restart () =
  match Queue.take_opt runnable with
  | None -> ()
  | Some f -> f ()

(** The scheduler. *)

let rec corun (f: unit -> unit) =
  match_with f ()
  { retc = (fun _ -> restart ());
    effc = (fun (type c) (eff: c Effect.t) ->
      match eff with
      | Spawn f -> Some (fun (k: (c,_) continuation) ->
                          suspend (fun () -> corun f); continue k ())
      | Yield  -> Some (fun (k: (c,_) continuation) ->
                          suspend (continue k); restart())
      | _ -> None);
    exnc = (fun exn -> raise exn)
  }

(** Example of use. *)

let task name n =
  for i = 1 to n do printf "%s%d " name i; yield() done

let _ =
  corun (fun () ->
    spawn (fun () -> task "a" 6);
    spawn (fun () -> task "b" 3);
    spawn (fun () -> task "c" 4))

(** Exercise: implement a "stop" function that terminates the calling
    coroutine.  Other corountines should keep running. *)

(** {1 Adding communication channels} *)

type 'a channel = {
    senders: ('a * (unit, unit) continuation) Queue.t;
    receivers: ('a, unit) continuation Queue.t
  }

type _ Effect.t +=
  | Send : 'a channel * 'a -> unit Effect.t
  | Recv : 'a channel -> 'a Effect.t

let new_channel () = { senders = Queue.create(); receivers = Queue.create() }

let send ch v = perform (Send(ch, v))

let recv ch = perform (Recv ch)

let rec corun (f: unit -> unit) =
  match_with f ()
  { retc = (fun _ -> restart ());
    effc = (fun (type c) (eff: c Effect.t) ->
      match eff with
      | Spawn f -> Some (fun (k: (c,_) continuation) ->
                          suspend (fun () -> corun f); continue k ())
      | Yield  -> Some (fun (k: (c,_) continuation) ->
                          suspend (continue k); restart())
      | Send(ch, v) -> Some (fun (k: (c,_) continuation) ->
          match Queue.take_opt ch.receivers with
          | Some rc -> suspend (continue k); continue rc v
          | None    -> Queue.add (v, k) ch.senders; restart())
      | Recv ch -> Some (fun (k: (c,_) continuation) ->
          match Queue.take_opt ch.senders with
          | Some(v, sn) -> suspend (continue sn); continue k v
          | None        -> Queue.add k ch.receivers; restart())
      | _ -> None);
    exnc = (fun exn -> raise exn)
  }

(** A simple test. *)

let _ =
  let ch = new_channel() in
  corun (fun () ->
    spawn (fun () -> send ch "a");
    spawn (fun () -> send ch "b");
    printf "%s\n" (recv ch);
    printf "%s\n" (recv ch))

(** A more involved example: Eratosthenes' sieve using a pipeline of filters. *)

let rec eratosthenes input =
  let p = recv input in
  printf "%d " p;
  let output = new_channel() in
  spawn (fun () -> eratosthenes output);
  while true do
    let n = recv input in
    if n mod p <> 0 then send output n
  done

let _ =
  corun (fun () ->
    let ints = new_channel() in
    spawn (fun () -> eratosthenes ints);
    for i = 2 to 1000 do send ints i done)

(** {1 Asynchronous I/O} *)

open Unix

type _ Effect.t +=
  | Wait_for_reading : file_descr -> unit Effect.t
  | Wait_for_writing : file_descr -> unit Effect.t

(** I/O operations, in non-blocking mode.  If the operation cannot progress,
    we suspend ourselves by performing the appropriate effect, then retry. *)

let rec recv fd buf pos len fl =
  try
    Unix.recv fd buf pos len fl
  with Unix_error((EAGAIN|EWOULDBLOCK), _, _) ->
    perform (Wait_for_reading fd);
    recv fd buf pos len fl

let rec send fd buf pos len fl =
  try
    Unix.send fd buf pos len fl
  with Unix_error((EAGAIN|EWOULDBLOCK), _, _) ->
    perform (Wait_for_writing fd);
    send fd buf pos len fl

(** Record the tasks waiting on a file descriptor, indexed by file descriptor *)

let waiting_reads  : (file_descr, (unit,unit) continuation) Hashtbl.t = Hashtbl.create 16
let waiting_writes : (file_descr, (unit,unit) continuation) Hashtbl.t = Hashtbl.create 16

(** Restart a task waiting on the given file descriptor. *)

let restart_io tbl fd =
  let k = Hashtbl.find tbl fd in
  Hashtbl.remove tbl fd;
  suspend (continue k)

(** The task and I/O scheduler. *)

let rec restart () =
  match Queue.take_opt runnable with
  | Some f -> f ()
  | None ->
      (* See if there are tasks blocked on I/O that we could restart *)
      let rd_fds = Hashtbl.fold (fun fd _ acc -> fd::acc) waiting_reads [] in
      let wr_fds = Hashtbl.fold (fun fd _ acc -> fd::acc) waiting_writes [] in
      if rd_fds = [] && wr_fds = [] then () else begin
        let rdy_rd_fds, rdy_wr_fds, _ = Unix.select rd_fds wr_fds [] (-1.) in
        List.iter (restart_io waiting_reads) rdy_rd_fds;
        List.iter (restart_io waiting_writes) rdy_wr_fds;
        restart()
      end

(** The main handler. *)

let rec corun (f: unit -> unit) =
  match_with f ()
  { retc = (fun _ -> restart ());
    effc = (fun (type c) (eff: c Effect.t) ->
      match eff with
      | Spawn f -> Some (fun (k: (c,_) continuation) ->
                          suspend (fun () -> corun f); continue k ())
      | Yield  -> Some (fun (k: (c,_) continuation) ->
                          suspend (continue k); restart())
      | Wait_for_reading fd ->
                  Some (fun (k: (c,_) continuation) ->
                          Hashtbl.add waiting_reads fd k;
                          restart())
      | Wait_for_writing fd ->
                  Some (fun (k: (c,_) continuation) ->
                          Hashtbl.add waiting_writes fd k;
                          restart())
      | _ -> None);
    exnc = (fun exn -> raise exn)
  }
